-------------------------------------------------------------------------------
-- Concurrent Haskell Debugger 
--   DebugMain Module
--     by Thomas Boettcher <thomas.boettcher@gmx.de>
-------------------------------------------------------------------------------

module CHD.DebugMain
  (   
    startChd
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import CHD.GuiInterface
import CHD.GuiWindow
import CHD.DebugState
import CHD.BaseTypes
import CHD.BaseFunctions
import CHD.Environment
import CHD.PriorDoubleChannel

import qualified Control.Concurrent as C
import Data.FiniteMap


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

startChd :: C.ThreadId -> IO ()
startChd myId = debugger (initDebuggerState myId)


-------------------------------------------------------------------------------
-- DEBUGGER message receiver (main)
--   receiving from Concurrent-functions as well as from Gui-Interface
-------------------------------------------------------------------------------

debugger :: DebuggerState -> IO ()
debugger state = do
  message <- readDblChan debugMsgChan
  case message of
    ---------------------------------------------------------------------------
    -- messages from Concurrent-functions (LESS-Channel)
    --   this is the main work for messages from the concurrent program
    ---------------------------------------------------------------------------
    Right (actorId, msg, stopper) -> do
      let actorNo = lookupThreadNo state actorId
      stopState <- stopThread state actorNo stopper
      transformState <- transformAndVisualize stopState actorNo msg
      breakState <- breakCheck transformState actorNo msg
      debugger breakState
    ---------------------------------------------------------------------------
    -- messages from Gui-Interface (PRIOR-Channel)
    --   just intercepting messages from the gui.
    --   Visualize was quite some work, since there could be the 
    --	   Visualize options-window open.
    ---------------------------------------------------------------------------
    Left msg -> do
      case msg of
	CHDQuit -> do
	  let ids = keysFM (threadMap state)
	  chdMessage("received Quit")
	  mapM_ (\threadId -> do
	          chdMessage("killing " ++ 
			     (show (lookupThreadNo state threadId)))
		  C.killThread threadId
--		  C.threadDelay 10000
		)
		ids
          myThreadId <- C.myThreadId
	  chdMessage("end")
	  C.killThread myThreadId
	Continue threadNo -> do
	  newState <- continueThread state threadNo 0
	  debugger newState
	AutoContinue threadNo unique -> do
	  newState <- continueThread state threadNo unique
	  debugger newState
	Visualize object switch -> do
	  let newState = stateSetVisualize state object switch
	  if switch 
	    then do
	      case object of
	        ThreadObj tNo -> do
		  let threadState = lookupThread newState tNo 
		      status = maybe (threadStatus threadState) 
				     (\_ -> TSDebug) 
				     (threadStopper threadState)
		  sendGuiMsg $ GuiAdd object 
			  [ FillColor (colorFct status),
			    Name (threadLabel threadState) ]
		  case threadSuspend threadState of
		    Nothing -> return ()
		    Just (comObj, action) -> 
		      visualizeObject newState object comObj action True
		_ -> do
		  let comObjState = lookupComObj newState object 
		      elements = map (\x -> showShortcut x) 
				     (comObjContents comObjState) 
		  sendGuiMsg $ GuiAdd object 
			  [ Elements elements,
			    Name (comObjLabel comObjState) ]
		  mapM_ 
		    (\threadNo -> 
		      case threadSuspend $ lookupThread newState threadNo of
		        Nothing -> return ()
			Just (comObj, action) ->
			  visualizeObject 
			    newState (ThreadObj threadNo) comObj action True)
		    (comObjSuspend comObjState)
	      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
		     $ InsertListbox EnableListbox [show object]
	      debugger newState
	    else do
	      sendPriorGuiInteract (GuiRemove object 0)
	      sendPriorGuiInteract $ InitGuiWin VisualizationConfigWindow
		       $ InsertListbox DisableListbox [show object]
	      continueState <- case object of
	        ThreadObj threadNo -> continueThread newState threadNo 0
		_ -> do
		  let suspend = comObjSuspend $ lookupComObj newState object
		  foldM
		    (\state threadNo -> continueThread state threadNo 0)
		    newState
		    suspend
	      debugger continueState
	BreakAction action time -> do
	  sendGuiMsg $ GuiWinMessage ActionBreakConfigWindow $
	    InsertListbox (case time of
	      EveryTime -> DisableListbox
	      _ -> EnableListbox) [show action]
	  debugger (stateSetActionBreak state action time)
	DisplayPosition treadName -> do
	  displayPosition state False $ parseObject treadName
	  debugger state
	GetInfoGuiWin VisualizationConfigWindow -> do
	  let list = map (\object -> (object, lookupVisualize state object))
			 (lookupObjectList state)
	      (elist, dlist) = foldr (\(key,bool) (enable, disable) -> 
				       if bool 
				         then (key:enable, disable) 
					 else (enable, key:disable)) 
				     ([],[]) 
				     list
	  sendPriorGuiInteract (InitGuiWin VisualizationConfigWindow 
		     (InsertListbox DisableListbox (map show dlist)))
	  sendPriorGuiInteract (InitGuiWin VisualizationConfigWindow 
		     (InsertListbox EnableListbox (map show elist)))
	  debugger state
	GetInfoGuiWin ActionBreakConfigWindow -> do
	  let list = map (\action -> (action, lookupActionBreak state action))
			 [minBound..maxBound]
	      (elist, dlist) = foldr (\(key,time) (enable, disable) ->
				       if (time == NoTime)
				         then (key:enable, disable)
					 else (enable, key:disable))
				     ([],[])
				     list
	  sendPriorGuiInteract (InitGuiWin ActionBreakConfigWindow
		     (InsertListbox DisableListbox (map show dlist)))
	  sendPriorGuiInteract (InitGuiWin ActionBreakConfigWindow
		     (InsertListbox EnableListbox (map show elist)))
	  debugger state
	GetInfoGuiWin SourceCodeViewWindow -> do
	  let objList = filter (lookupVisualize state) 
			       (lookupThreadObjectList state)
	      stringList = map show objList
	  sendPriorGuiInteract $ InitGuiWin SourceCodeViewWindow
		     $ InsertListbox SelectListbox stringList
	  displayPosition state False $ last objList
	  debugger state
	GetInfoGuiWin ProgressViewWindow -> do
	  let objList = filter (lookupVisualize state)
			       (lookupThreadObjectList state)
	  mapM_ (\obj -> sendPriorGuiInteract $ InitGuiWin ProgressViewWindow
		     $ InsertCanvas ProgressCanvas obj 
				    (lookupProgress state obj)) objList
	_ -> do
	  chdMessage("received unknown Message (prior):")
	  chdMessage("     '" ++ (show msg) ++ "'")
	  debugger state		


-------------------------------------------------------------------------------
-- operations THREAD-section
-------------------------------------------------------------------------------

-- this is invoked by a message from the GUI. 
-- it checks the identifier and continues the thread with the right number.

continueThread :: DebuggerState -> ThreadNo -> Int -> IO DebuggerState
continueThread state threadNo identifier = do
  let threadObj = ThreadObj threadNo
      threadState = lookupThread state threadNo
  case (threadStopper threadState) of
    Just (stopperId, stopper) -> 
      if ((identifier == 0) || (identifier == stopperId))
	then do 
	  C.putMVar stopper ()
	  if (lookupVisualize state threadObj)
	    then sendGuiMsg $ GuiPar threadObj
			[FillColor (colorFct
				 (threadStatus $ lookupThread state threadNo))]
	    else return ()
	  return (stateRemoveThreadStopper state threadNo)
	else do
	  return state
    _ -> return state


-- simply adds the mvar which a thread suspends on.

stopThread :: DebuggerState -> ThreadNo -> C.MVar () -> IO DebuggerState
stopThread state threadNo stopper = do
  return (stateAddThreadStopper state threadNo stopper)


{-
-- this is substituted due to the concept of sending a message to the gui to
--   return and say it is okay to continue
--   stopperId seems not to be needed anymore
--   verify?

autoContinueThread :: DebuggerState -> ThreadNo -> IO ()
autoContinueThread state threadNo = do
  maybe (return ())
	(\(stopperId, _) -> do
	  C.forkIO (do C.threadDelay ((autoContinueTime state) * 1000)
		       sendInteractMsg (AutoContinue threadNo stopperId))
	  return ())
	(threadStopper $ lookupThread state threadNo)
-}


-------------------------------------------------------------------------------
-- State Transformer & visualizer
-------------------------------------------------------------------------------

-- main work for the concurrent program: 
--   here every message is split up and checked what should be changed 
--     in the state
--   it controls either when the visualization-message should be send.

transformAndVisualize :: DebuggerState -> ThreadNo -> DebugMsg 
			 -> IO DebuggerState
transformAndVisualize startState actorNo msg = do
  let state = stateRemoveSuspend startState actorNo
      actorObj = ThreadObj actorNo
      visualizeActor = lookupVisualize state actorObj

-- submit color to Gui if needed.
  if visualizeActor 
    then do 
      let actionMsg = case msg of 
	    ThreadBreak cause -> show (reduceDebugMsg msg) ++ " (" ++ cause 
				 ++ ")"
	    _ -> show (reduceDebugMsg msg)
      sendGuiMsg (GuiPar actorObj [ -- FillColor (colorFct TSDebug),
				Action actionMsg ])
    else return ()

  case msg of

-- programlabel: submit it to Gui
    ProgramStart (Identity newId) -> do
      let (addState, threadNo) = stateAddThread state newId
	  threadObj = ThreadObj threadNo
          threadState = (lookupThread state threadNo) {
	    threadStatus = TSRunning,
	    threadParent = actorNo,
	    threadBreak = NoTime
	    }
	  statState = stateSetThread addState threadNo threadState
	  newState' = stateAddStep statState actorObj 
				   (NoPosition,threadObj,TAFork,False)
      visualizeObject newState' actorObj threadObj TAFork False
      return newState'
    ProgramLabel name -> do
      sendGuiMsg (GuiTitle name)
      return state
	-----------------------------------------------------------------------
	-- messages from CHD.ConcurrentMain
	-----------------------------------------------------------------------

-- fork: add new thread to state-data and submit to Gui
    ThreadForkSuspend (Identity newId) codePos -> do
      let (addState, threadNo) = stateAddThread state newId
	  threadObj = ThreadObj threadNo
	  newState = stateSetSuspend addState actorNo threadObj TAFork
	  newState' = stateAddStep newState actorObj 
				   (codePos,threadObj,TAFork,True)
      visualizeObject newState' actorObj threadObj TAFork True
      return newState'
    ThreadFork (Identity threadId) (CHD mvar) codePos -> do
      let threadNo = lookupThreadNo state threadId
	  threadObj = ThreadObj threadNo
	  threadState = (lookupThread state threadNo) {
	    threadStatus = TSRunning,
	    threadParent = actorNo,
	    threadBreak = NoTime
-- (if (autoContinueOption == 1) then NoTime else EveryTime) 
	    }
	  statState = stateSetThread state threadNo threadState
	  newState' = stateAddStep statState actorObj 
				   (codePos,threadObj,TAFork,False)
      visualizeObject newState' actorObj threadObj TAFork False
      sendGuiMsg $ GuiWinMessage ProgressViewWindow 
	       $ InsertCanvas ProgressCanvas actorObj 
			      [(codePos,threadObj,TAFork,False)]
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
	       $ InsertListbox EnableListbox [show threadObj]
      C.putMVar mvar ()
      return newState'
    ThreadKillSuspend (Identity threadId) codePos -> do
      let threadNo = lookupThreadNo state threadId
	  threadObj = ThreadObj threadNo
	  newState' = stateAddStep state actorObj 
				   (codePos,threadObj,TAKill,True)
      visualizeObject newState' actorObj threadObj TAKill True  
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    ThreadKill (Identity threadId) codePos -> do
      let threadNo = lookupThreadNo state threadId
	  threadObj = ThreadObj threadNo
	  newState' = stateAddStep state actorObj 
				   (codePos,threadObj,TAKill,False)
      visualizeObject newState' actorObj threadObj TAKill False
      if (lookupVisualize newState' threadObj) 
        then do
	  sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		 $ RemoveListbox EnableListbox [show threadObj]
	  sendGuiMsg (GuiDel threadObj)
	else sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		    $ RemoveListbox DisableListbox [show threadObj]
--      let (newState', threadNo) = stateRemoveThread newState threadId
      return (stateSetThreadStatus newState' threadNo TSKilled)
    ThreadStart ->
      return (stateSetThreadStatus state actorNo TSSuspend)
    ThreadEnd -> do
      if visualizeActor 
        then do
	  sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		 $ RemoveListbox EnableListbox [show actorObj]
	  sendGuiMsg (GuiDel actorObj)
	else sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		    $ RemoveListbox DisableListbox [show actorObj]
--      let (newState, threadNo) = stateRemoveThread state actorId
      return (stateSetThreadStatus state actorNo TSFinished)
    ThreadDelaySuspend -> do
      return (stateSetThreadStatus state actorNo TSDelay)
    ThreadDelay -> do
      return (stateSetThreadStatus state actorNo TSRunning)
    ThreadLabel name oldName -> do
      C.putMVar oldName (lookupLabel state actorObj)
      let label = (if (name == "") 
            then (show actorNo) 
	    else ((showShortcut actorNo) ++ ":" ++ name))
      if visualizeActor
        then sendGuiMsg $ GuiPar actorObj [Name label]
	else return ()
      return $ stateSetLabel state actorObj label 
    ThreadMark (Identity threadId) color -> do    
      let threadNo = lookupThreadNo state threadId
	  threadObj = ThreadObj threadNo
      sendGuiMsg $ GuiPar threadObj [Thickness 2,OutlineColor color]
      return state
    ThreadUnmark (Identity threadId) -> do    
      let threadNo = lookupThreadNo state threadId
	  threadObj = ThreadObj threadNo
      sendGuiMsg $ GuiPar threadObj [Thickness 1,OutlineColor "black"]
      return state


----------------------------------------------------------------------
-- messages from CHD.ConcurrentMVar
----------------------------------------------------------------------

    MVarNewEmptySuspend (CHD returnNewNoMVar) codePos -> do
      let (addState, mvarNo) = stateAddMVar state
	  mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend addState actorNo mvarObj TACreate
	  newState' = stateAddStep newState actorObj 
				   (codePos,NoObject,TACreate,True)
      C.putMVar returnNewNoMVar mvarNo
      visualizeObject newState' actorObj mvarObj TACreate True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    MVarNewEmpty mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState' = stateAddStep state actorObj 
				   (codePos,mvarObj,TACreate,False)
      visualizeObject newState' actorObj mvarObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show mvarObj]          
      return (stateSetThreadStatus newState' actorNo TSRunning)
    MVarNewSuspend (CHD returnNewNoMVar) codePos -> do
      let (addState, mvarNo) = stateAddMVar state
	  mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend addState actorNo mvarObj TACreate
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TACreate,True)
      C.putMVar returnNewNoMVar mvarNo
      visualizeObject newState' actorObj mvarObj TACreate True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    MVarNew mvarNo label codePos -> do
      let mvarObj = MVarObj mvarNo
          newState = stateSetContentLabel state mvarObj [actorNo] [label]
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TACreate,False)
      visualizeObject newState' actorObj mvarObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show mvarObj]
      return (stateSetThreadStatus newState' actorNo TSRunning)
    MVarTakeSuspend mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend state actorNo mvarObj TARead
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TARead,True)
      visualizeObject newState' actorObj mvarObj TARead True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    MVarTake mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetContentLabel state mvarObj [] []
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TARead,False)
      visualizeObject newState' actorObj mvarObj TARead False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    MVarReadSuspend mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend state actorNo mvarObj TARead
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TARead,True)
      visualizeObject newState' actorObj mvarObj TARead True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    MVarRead mvarNo codePos -> do
      let mvarObj = (MVarObj mvarNo)
	  newState' = stateAddStep state actorObj 
				   (codePos,mvarObj,TARead,False)
      visualizeObject newState' actorObj mvarObj TARead False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    MVarPutSuspend mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend state actorNo mvarObj TAWrite
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TAWrite,True)
      visualizeObject newState' actorObj mvarObj TAWrite True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    MVarPut mvarNo label codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetContentLabel state mvarObj [actorNo] [label]
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TAWrite,False)
      visualizeObject newState' actorObj mvarObj TAWrite False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    MVarSwapSuspend mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend state actorNo mvarObj TASwap
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TASwap,True)
      visualizeObject newState' actorObj mvarObj TASwap True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    MVarSwap mvarNo _ codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetContentLabel state mvarObj [actorNo] []
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TASwap,False)
      visualizeObject newState' actorObj mvarObj TASwap False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    MVarTryTake mvarNo codePos -> do
      let mvarObj = MVarObj mvarNo
	  newState = stateSetSuspend state actorNo mvarObj TATry
	  newState' = stateAddStep newState actorObj 
				   (codePos,mvarObj,TATry,True)
      visualizeObject newState' actorObj mvarObj TATry True
      return newState'
    MVarTry mvarNo codePos -> do
      visualizeObject state actorObj (MVarObj mvarNo) TATry False
      return state
    MVarDied mvarNo -> do
      let mvarObj = MVarObj mvarNo 
      if lookupVisualize state mvarObj
        then do
	  sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		 $ RemoveListbox EnableListbox [show mvarObj]
	  sendGuiMsg (GuiDel mvarObj)
	else sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		    $ RemoveListbox DisableListbox [show mvarObj]
      return (stateRemoveComObj state mvarObj)
    MVarLabel mvarNo name oldName -> do
      let mvarObj = MVarObj mvarNo 
      C.putMVar oldName (lookupLabel state mvarObj)
      let label = (if (name == "") 
            then (show mvarNo)
            else ((showShortcut mvarNo) ++ ":" ++ name))
      if lookupVisualize state mvarObj
        then sendGuiMsg $ GuiAdd mvarObj [Name label]
	else return ()
      return $ stateSetLabel state mvarObj label


-----------------------------------------------------------------------
-- messages from CHD.ConcurrentChannel
-----------------------------------------------------------------------
    
    ChanNewSuspend (CHD chanNoMVar) codePos -> do
      let (newState, chanNo) = stateAddChannel state
	  chanObj = ChanObj chanNo
	  newState' = stateSetSuspend newState actorNo chanObj TACreate
      visualizeObject newState actorObj chanObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show chanObj]
      C.putMVar chanNoMVar chanNo
      return newState'
    ChanNew chanNo codePos -> do
      let chanObj = ChanObj chanNo
      visualizeObject state actorObj chanObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show chanObj]
      return state
    ChanWriteSuspend chanNo codePos -> do
      let chanObj = ChanObj chanNo
	  newState = stateSetSuspend state actorNo chanObj TAWrite
	  newState' = stateAddStep newState actorObj 
				   (codePos,chanObj,TAWrite,True)
      visualizeObject newState' actorObj chanObj TAWrite True
      return newState'
    ChanWrite chanNo label codePos -> do
      let chanObj = ChanObj chanNo
	  newState = stateChannelAddMessageLabel state chanNo actorNo [label]
      visualizeObject newState actorObj chanObj TAWrite False
      return newState  
    ChanReadSuspend chanNo codePos -> do
      let chanObj = ChanObj chanNo
	  newState = stateSetSuspend state actorNo chanObj TARead
      visualizeObject newState actorObj chanObj TARead True
      return (stateSetThreadStatus newState actorNo TSSuspend)
    ChanRead chanNo codePos -> do
      let newState = stateChannelRemoveMessage state chanNo
      visualizeObject newState actorObj (ChanObj chanNo) TARead False
      return (stateSetThreadStatus newState actorNo TSRunning)
    ChanDied chanNo -> do
      let chanObj = ChanObj chanNo
      if lookupVisualize state chanObj
        then do
	  sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		 $ RemoveListbox EnableListbox [show chanObj]
	  sendGuiMsg (GuiDel chanObj)
	else sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		    $ RemoveListbox DisableListbox [show chanObj]
      return (stateRemoveComObj state (ChanObj chanNo))
    ChanLabel chanNo name oldName -> do
      let chanObj = ChanObj chanNo
      C.putMVar oldName (lookupLabel state chanObj)
      let label = (if (name == "")
            then (show chanNo)
	    else ((showShortcut chanNo) ++ ":" ++ name))
      if lookupVisualize state chanObj
        then sendGuiMsg $ GuiAdd chanObj [Name label]
	else return ()
      return $ stateSetLabel state chanObj label


-----------------------------------------------------------------------
-- messages from CHD.ConcurrentQSem
-----------------------------------------------------------------------
    
    QSemNewSuspend (CHD returnNewNoMVar) quantity codePos -> do
      let (addState, qsemNo) = stateAddQSem state quantity
	  qsemObj = QSemObj qsemNo
	  newState = stateSetSuspend addState actorNo qsemObj TACreate
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemObj,TACreate,True)
      C.putMVar returnNewNoMVar qsemNo
      visualizeObject newState' actorObj qsemObj TACreate True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    QSemNew qsemNo codePos -> do
      let qsemObj = QSemObj qsemNo
	  newState' = stateAddStep state actorObj 
				   (codePos,qsemObj,TACreate,False)
      visualizeObject newState' actorObj qsemObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show qsemObj]
      return (stateSetThreadStatus newState' actorNo TSRunning)
    QSemWaitSuspend qsemNo codePos -> do
      let qsemObj = QSemObj qsemNo
	  newState = stateSetSuspend state actorNo qsemObj TARead
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemObj,TARead,True)
      visualizeObject newState' actorObj qsemObj TARead True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    QSemWait qsemNo codePos -> do
      let qsemObj = QSemObj qsemNo
	  newState = stateDecQuantity state qsemObj
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemObj,TARead,False)
      visualizeObject newState' actorObj qsemObj TARead False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    QSemSignalSuspend qsemNo codePos -> do
      let qsemObj = QSemObj qsemNo
	  newState = stateSetSuspend state actorNo qsemObj TAWrite
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemObj,TAWrite,True)
      visualizeObject newState' actorObj qsemObj TAWrite True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    QSemSignal qsemNo codePos -> do
      let qsemObj = QSemObj qsemNo
	  newState = stateIncQuantity state qsemObj
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemObj,TAWrite,False)
      visualizeObject newState' actorObj qsemObj TAWrite False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    QSemLabel qsemNo name oldName -> do
      let qsemObj = QSemObj qsemNo 
      C.putMVar oldName (lookupLabel state qsemObj)
      let label = (if (name == "") 
            then (show qsemNo)
            else ((showShortcut qsemNo) ++ ":" ++ name))
      if lookupVisualize state qsemObj
        then sendGuiMsg $ GuiAdd qsemObj [Name label]
	else return ()
      return $ stateSetLabel state qsemObj label


-----------------------------------------------------------------------
-- messages from CHD.ConcurrentQSemN
-----------------------------------------------------------------------

    QSemNNewSuspend (CHD returnNewNoMVar) quantity codePos -> do
      let (addState, qsemnNo) = stateAddQSemN state quantity
	  qsemnObj = QSemNObj qsemnNo
	  newState = stateSetSuspend addState actorNo qsemnObj TACreate
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemnObj,TACreate,True)
      C.putMVar returnNewNoMVar qsemnNo
      visualizeObject newState' actorObj qsemnObj TACreate True
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show qsemnObj]
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    QSemNNew qsemnNo codePos -> do
      let qsemnObj = QSemNObj qsemnNo
	  newState' = stateAddStep state actorObj 
				   (codePos,qsemnObj,TACreate,False)
      visualizeObject newState' actorObj qsemnObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show qsemnObj]
      return (stateSetThreadStatus newState' actorNo TSRunning)
    QSemNWaitSuspend qsemnNo codePos -> do
      let qsemnObj = QSemNObj qsemnNo
	  newState = stateSetSuspend state actorNo qsemnObj TARead
	  newState' = stateAddStep newState actorObj
				   (codePos,qsemnObj,TARead,True)
      visualizeObject newState' actorObj qsemnObj TARead True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    QSemNWait quantity qsemnNo codePos -> do
      let qsemnObj = QSemNObj qsemnNo
          newState = stateDecQuantityN state qsemnObj quantity
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemnObj,TARead,False)
      visualizeObject newState' actorObj qsemnObj TARead False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    QSemNSignalSuspend qsemnNo codePos -> do
      let qsemnObj = QSemNObj qsemnNo
	  newState = stateSetSuspend state actorNo qsemnObj TAWrite
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemnObj,TAWrite,True)
      visualizeObject newState' actorObj qsemnObj TAWrite True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    QSemNSignal quantity qsemnNo codePos -> do
      let qsemnObj = QSemNObj qsemnNo
	  newState = stateIncQuantityN state qsemnObj quantity
	  newState' = stateAddStep newState actorObj 
				   (codePos,qsemnObj,TAWrite,False)
      visualizeObject newState' actorObj qsemnObj TAWrite False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    QSemNLabel qsemnNo name oldName -> do
      let qsemnObj = QSemNObj qsemnNo 
      C.putMVar oldName (lookupLabel state qsemnObj)
      let label = (if (name == "") 
            then (show qsemnNo)
            else ((showShortcut qsemnNo) ++ ":" ++ name))
      if lookupVisualize state qsemnObj
        then sendGuiMsg $ GuiAdd qsemnObj [Name label]
	else return ()
      return $ stateSetLabel state qsemnObj label
   

-----------------------------------------------------------------------
-- messages from CHD.ConcurrentSampleVar
-----------------------------------------------------------------------
    
    SampleVarNewEmptySuspend (CHD returnNewNoMVar) codePos -> do
      let (addState, samplevarNo) = stateAddSampleVar state
	  samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetSuspend addState actorNo samplevarObj TACreate
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TACreate,True)
      C.putMVar returnNewNoMVar samplevarNo
      visualizeObject newState' actorObj samplevarObj TACreate True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    SampleVarNewEmpty samplevarNo codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState' = stateAddStep state actorObj 
				   (codePos,samplevarObj,TACreate,False)
      visualizeObject newState' actorObj samplevarObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show samplevarObj]
      return (stateSetThreadStatus newState' actorNo TSRunning)
    SampleVarNewSuspend (CHD returnNewNoMVar) codePos -> do
      let (addState, samplevarNo) = stateAddSampleVar state
	  samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetSuspend addState actorNo samplevarObj TACreate
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TACreate,True)
      C.putMVar returnNewNoMVar samplevarNo
      visualizeObject newState actorObj samplevarObj TACreate True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    SampleVarNew samplevarNo label codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetContentLabel state samplevarObj [actorNo] [label]
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TACreate,False)
      visualizeObject newState actorObj samplevarObj TACreate False
      sendGuiMsg $ GuiWinMessage VisualizationConfigWindow
	     $ InsertListbox EnableListbox [show samplevarObj]
      return (stateSetThreadStatus newState' actorNo TSRunning)
    SampleVarWriteSuspend samplevarNo codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetSuspend state actorNo samplevarObj TAWrite
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TAWrite,True)
      visualizeObject newState actorObj samplevarObj TAWrite True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    SampleVarWrite samplevarNo label codePos -> do
      let samplevarObj = (SampleVarObj samplevarNo)
	  newState = stateSetContentLabel state samplevarObj [actorNo] [label]
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TAWrite,False)
      visualizeObject newState' actorObj samplevarObj TAWrite False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    SampleVarReadSuspend samplevarNo codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetSuspend state actorNo samplevarObj TARead
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TARead,True)
      visualizeObject newState' actorObj samplevarObj TARead True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    SampleVarRead samplevarNo codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetContentLabel state samplevarObj [] []
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TARead,False)
      visualizeObject newState' actorObj samplevarObj TARead False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    SampleVarEmptySuspend samplevarNo codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetSuspend state actorNo samplevarObj TAWrite
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TAWrite,True)
      visualizeObject newState' actorObj samplevarObj TAWrite True
      return (stateSetThreadStatus newState' actorNo TSSuspend)
    SampleVarEmpty samplevarNo codePos -> do
      let samplevarObj = SampleVarObj samplevarNo
	  newState = stateSetContentLabel state samplevarObj [] []
	  newState' = stateAddStep newState actorObj 
				   (codePos,samplevarObj,TAWrite,False)
      visualizeObject newState' actorObj samplevarObj TAWrite False
      return (stateSetThreadStatus newState' actorNo TSRunning)
    SampleVarDied samplevarNo -> do
      let samplevarObj = SampleVarObj samplevarNo 
      if lookupVisualize state samplevarObj
        then do
	  sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		 $ RemoveListbox EnableListbox [show samplevarObj]
	  sendGuiMsg (GuiDel samplevarObj)
	else sendGuiMsg $ GuiWinMessage VisualizationConfigWindow 
		    $ RemoveListbox DisableListbox [show samplevarObj]
      return (stateRemoveComObj state samplevarObj)
    SampleVarLabel samplevarNo name oldName -> do
      let samplevarObj = SampleVarObj samplevarNo 
      C.putMVar oldName (lookupLabel state samplevarObj)
      let label = (if (name == "") 
            then (show samplevarNo)
            else ((showShortcut samplevarNo) ++ ":" ++ name))
      if lookupVisualize state samplevarObj
        then sendGuiMsg $ GuiAdd samplevarObj [Name label]
	else return ()
      return $ stateSetLabel state samplevarObj label   

    _ -> return state


-------------------------------------------------------------------------------
-- Visualize Action
-------------------------------------------------------------------------------

displayPosition ::  DebuggerState -> Bool -> Object -> IO ()
displayPosition state insert actorObj = do
  let progress = lookupProgress state actorObj
  if length progress > 1
     then let (lastPos,_,_,lastSuspend) = head (drop 1 progress) in
	  case lastPos of
	       CodePosition _ lastLine ->
		   sendGuiMsg $ GuiWinMessage SourceCodeViewWindow
			      $ UnmarkEdit DisplayEdit lastLine
	       _ -> return ()
     else return ()
  if length progress > 0
     then do
	  let (pos,comObj,action,suspend) = head progress 
	      color = colorFct action
	  case (action,suspend,insert) of
	       (TAFork,False,True) ->
		   sendGuiMsg $ GuiWinMessage SourceCodeViewWindow
			     $ InsertListbox SelectListbox [show comObj]
	       (TAKill,False,True) ->
		   sendGuiMsg $ GuiWinMessage SourceCodeViewWindow
			     $ RemoveListbox SelectListbox [show comObj]
	       _ -> return ()
	  case pos of
	       CodePosition file line ->
		   do
		   sendGuiMsg $ GuiWinMessage SourceCodeViewWindow
			      $ Select SelectListbox file (show actorObj)
		   sendGuiMsg $ GuiWinMessage SourceCodeViewWindow
			      $ MarkEdit DisplayEdit line color "white"
	       _ -> return ()
     else return ()


-- checks what should be visualized and what message should be send to GUI

-- visualizeObject state actorObj comobj action suspend?
visualizeObject :: DebuggerState -> Object -> Object -> ThreadAction -> Bool 
				 -> IO ()
visualizeObject state actorObj comObj action suspend =
  let visualizeComObj = lookupVisualize state comObj
      visualizeActorObj = lookupVisualize state actorObj
      arrow = case action of
        TARead   -> ArrowObj (comObj, actorObj)
	TAWrite  -> ArrowObj (actorObj, comObj)
	TAFork	 -> ArrowObj (actorObj, comObj)
	_	 -> ArrowObj (actorObj, comObj)
      parameter = case (action, actorObj) of
        (TACreate, ThreadObj actorNo) -> [Parent actorNo]
	(TAFork, ThreadObj actorNo)   -> [Parent actorNo]
	_			      -> []
      color = colorFct comObj
      colorPar = case comObj of
	QSemObj  _ -> case quantity of
			   (Quantity (x+1)) -> [FillColor color,
					        OutlineColor color]
			   (Quantity _)     -> [FillColor "DarkRed",
					        OutlineColor "DarkRed"]
	QSemNObj _ -> case quantity of
			   (Quantity (x+1)) -> [FillColor color,
					        OutlineColor color]
			   (Quantity _)     -> [FillColor "red3",
					        OutlineColor "red3"]
	_	   -> []
      contentLabel = lookupContentLabel state comObj
      quantity = lookupQuantity state comObj
      elements = case comObj of
	QSemObj  _ -> [show quantity]
	QSemNObj _ -> [show quantity]
	_ -> if( contentLabel == [] || contentLabel == [""] )
	     then (map (\x -> showShortcut x) (lookupContents state comObj))
	     else (zipWith (\x y-> y ++ " : " ++ x) contentLabel
	     	  (map (\x -> showShortcut x) (lookupContents state
		  comObj))) in
  do
  if visualizeComObj
    then do
      displayPosition state True actorObj
      if suspend
        then return () -- sendGuiMsg (GuiAdd comObj parameter) : is eh leer.
	else if action==TAFork
	     then sendGuiMsg (GuiAdd comObj (parameter ++ colorPar))
	     else sendGuiMsg (GuiAdd comObj (((Elements elements):parameter)
					     ++ colorPar))		     
      if visualizeActorObj
        then if suspend
	     then if action==TAFork || action==TACreate
		  then return ()
		  else sendGuiMsg (GuiAdd arrow [ Thickness 1,
					FillColor (colorFct action) ])
	     else do
	          sendGuiMsg (GuiAdd arrow
			     [ Thickness 2,
			       FillColor (colorFct action) ])
	          sendGuiMsg (GuiDel arrow)
	else return ()
    else return ()


-------------------------------------------------------------------------------
-- Break Check
-------------------------------------------------------------------------------

-- checks like described in diploma thesis wether a thread should 
--    automaticly continue or not.

breakCheck :: DebuggerState -> ThreadNo -> DebugMsg -> IO DebuggerState
breakCheck state actorNo msg = do
  element <- secondObject state msg
  let (visualizeElement, breakElement) = case element of
        NoObject -> (True, NoTime)
	_ -> (lookupVisualize state element, lookupBreak state element)
      actorObj = ThreadObj actorNo
      visualizeActor = lookupVisualize state actorObj
      breakActor = lookupBreak state actorObj
      breakAction = lookupActionBreak state (reduceDebugMsg msg)
      continue = and [ (breakActor > EveryTime),
		       (breakAction > EveryTime),
		       (breakElement > EveryTime) ]
--      newState = stateSetBreak state actorObj (pred breakActor)
  if (continue || (not visualizeActor) || (not visualizeElement))
    then sendGuiMsg (ThreadReady actorNo)
    else sendGuiMsg (GuiPar actorObj [ FillColor (colorFct TSDebug) ])
  {- case msg of
    ThreadFork (Identity newId) _ -> do
      let threadNo = lookupThreadNo state newId
          threadObj = ThreadObj threadNo
	  visualizeElement = lookupVisualize state threadObj
	  breakElement = lookupBreak state threadObj
	  breakAction = lookupActionBreak state ThreadStartA
	  continueElement = and [ (breakElement > EveryTime),
	  			  (breakAction > EveryTime) ]
      if (continueElement || (not visualizeElement))
        then sendGuiMsg (ThreadReady threadNo)
	else sendGuiMsg (GuiPar threadObj [ FillColor (colorFct TSDelay) ])
    _ -> return () -}
  return state


-------------------------------------------------------------------------------
-- second Object of DebugMsg
-------------------------------------------------------------------------------

-- just returns the second object for the check.

secondObject :: DebuggerState -> DebugMsg -> IO Object
secondObject state msg =
  case msg of
    ThreadForkSuspend (Identity threadId) _ -> 
      return (ThreadObj (lookupThreadNo state threadId))
    ThreadKillSuspend (Identity threadId) _ -> 
      return (ThreadObj (lookupThreadNo state threadId))
    ThreadKill (Identity threadId) _ -> 
      return (ThreadObj (lookupThreadNo state threadId))
    MVarNewEmptySuspend (CHD mvarNoMVar) _ -> do
      mvarNo <- C.readMVar mvarNoMVar
      return (MVarObj mvarNo)
    MVarNewEmpty mvarNo _ -> do
      return (MVarObj mvarNo)
    MVarNew mvarNo _ _ -> do
      return (MVarObj mvarNo)
    MVarTakeSuspend mvarNo _ -> return (MVarObj mvarNo)
    MVarTake mvarNo _ -> return (MVarObj mvarNo)
    MVarReadSuspend mvarNo _ -> return (MVarObj mvarNo)
    MVarRead mvarNo _ -> return (MVarObj mvarNo)
    MVarPutSuspend mvarNo _ -> return (MVarObj mvarNo)
    MVarPut mvarNo _ _ -> return (MVarObj mvarNo)
    MVarSwapSuspend mvarNo _ -> return (MVarObj mvarNo)
    MVarSwap mvarNo _ _ -> return (MVarObj mvarNo)
    MVarTryTake mvarNo _ -> return (MVarObj mvarNo)
    MVarTry mvarNo _ -> return (MVarObj mvarNo)
    MVarDied mvarNo -> return (MVarObj mvarNo)
    MVarLabel mvarNo _ _ -> return (MVarObj mvarNo)

    ChanNewSuspend (CHD chanNoMVar) _ -> do
      chanNo <- C.readMVar chanNoMVar
      return (ChanObj chanNo)
    ChanNew chanNo _ -> do
      return (ChanObj chanNo)
    ChanWriteSuspend chanNo _ -> return (ChanObj chanNo) 
    ChanWrite chanNo _ _ -> return (ChanObj chanNo) 
    ChanReadSuspend chanNo _ -> return (ChanObj chanNo) 
    ChanRead chanNo _ -> return (ChanObj chanNo) 
    ChanUnGetSuspend chanNo -> return (ChanObj chanNo) 
    ChanUnGet chanNo -> return (ChanObj chanNo) 
    ChanDied chanNo -> return (ChanObj chanNo) 
    ChanLabel chanNo _ _ -> return (ChanObj chanNo) 

    QSemNewSuspend (CHD qsemNoMVar) _ _ -> do
      qsemNo <- C.readMVar qsemNoMVar
      return (QSemObj qsemNo)
    QSemNew qsemNo _ -> do
      return (QSemObj qsemNo)  
    QSemWaitSuspend qsemNo _ -> return (QSemObj qsemNo)
    QSemWait qsemNo _ -> return (QSemObj qsemNo)
    QSemSignalSuspend qsemNo _ -> return (QSemObj qsemNo)
    QSemSignal qsemNo _ -> return (QSemObj qsemNo)
    QSemDied qsemNo -> return (QSemObj qsemNo)
    QSemLabel qsemNo _ _ -> return (QSemObj qsemNo)

    QSemNNewSuspend (CHD qsemnNoMVar) _ _ -> do
      qsemnNo <- C.readMVar qsemnNoMVar
      return (QSemNObj qsemnNo)
    QSemNNew qsemnNo _ -> do
      return (QSemNObj qsemnNo)
    QSemNWaitSuspend qsemnNo _ -> return (QSemNObj qsemnNo)
    QSemNWait _ qsemnNo _ -> return (QSemNObj qsemnNo)
    QSemNSignalSuspend qsemnNo _ -> return (QSemNObj qsemnNo)
    QSemNSignal _ qsemnNo _ -> return (QSemNObj qsemnNo)
    QSemNDied qsemnNo -> return (QSemNObj qsemnNo)
    QSemNLabel qsemnNo _ _ -> return (QSemNObj qsemnNo)

    SampleVarNewEmptySuspend (CHD samplevarNoMVar) _ -> do
      samplevarNo <- C.readMVar samplevarNoMVar
      return (SampleVarObj samplevarNo)
    SampleVarNewEmpty samplevarNo _ -> do
      return (SampleVarObj samplevarNo)
    SampleVarNewSuspend (CHD samplevarNoMVar) _ -> do
      samplevarNo <- C.readMVar samplevarNoMVar
      return (SampleVarObj samplevarNo)
    SampleVarNew samplevarNo _ _ -> do
      return (SampleVarObj samplevarNo)
    SampleVarWriteSuspend samplevarNo _ -> return (SampleVarObj samplevarNo)
    SampleVarWrite samplevarNo _ _ -> return (SampleVarObj samplevarNo)
    SampleVarReadSuspend samplevarNo _ -> return (SampleVarObj samplevarNo)
    SampleVarRead samplevarNo _ -> return (SampleVarObj samplevarNo)
    SampleVarEmptySuspend samplevarNo _ -> return (SampleVarObj samplevarNo)
    SampleVarEmpty samplevarNo _ -> return (SampleVarObj samplevarNo)
    SampleVarDied samplevarNo -> return (SampleVarObj samplevarNo)
    SampleVarLabel samplevarNo _ _ -> return (SampleVarObj samplevarNo)
    
    _ -> return NoObject



    
